/***************************************************************************
 *            lan_simulator.c
 *
 *  Tue Oct  4 10:08:19 2005
 *  Copyright  2005  Francesco Carbone
 *  trini4mail@yahoo.it
 *	franci.carbone@gmail.com
 ****************************************************************************/

 
/* 	
 *	This code implements a program simulating a Bridged LAN for testing 
 *	the spanning tree protocol. In order to permit communication of multiple 
 *	virtual bridges executed on Freebsd machine, internet DGRAM type sockets 
 *	are used.
 *	It use a unique socket to receive BPDUs from any other	Bridge simulated
 *	and a multiple sockets to transmit BPDUs to every Bridge on simulated lans.
 *	The informations stored in the  Virtual Bridge configuration files
 *	are used to calculate the number of lans and the Bridged Lan topology.
 *	After initial topology learning, the process is ready to receive 
 *	BPDUs from Bridges running on the same machine or in another machine
 *	in the network.
 *
 *	Parameters convoyed in BPDU packet as port identifier ( portId ) 
 * 	and information stored in configuration file correlated to each bridge 
 *	are used  to :
 *		1) 	Identify bridge and port BPDU incoming from
 *		2)	Identify by configuration file, lans where BPDU received 
 *			will be trasnmitted
 *		3)	Identify for each lan which Bridge are attached
 *		4)	Calculate socket address for each virtual bridge 
 *			that will be simulated
 *
 *	Configuration file format :
 *
 *	Bridge-Address Bridge-priority maxAge helloTime forwardDelayTime numOfPorts	
 *	Port-number LAN-connected Port-priority pathCost 							
 *		'			'			'		'					
 *		'			'			'		'						
 *	Port-number LAN-connected Port-priority pathCost
 *
 *	In the configuration file there are specified lans to which 
 *	Bridge ports are attached. This informations are used
 *	to calculate the topology of the bridged lan. because 
 *	all information needed are known.
 *
 *	THINGS TO REMEMBER
 *		- 	if 10 Bridges will be simulated , configuration file 
 *			from bridge.conf1 to bridge.conf10 must be in 
 *			STPFILE directory
 *	
 */	
	
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <pthread.h>
#include <unistd.h>
#include <string.h>
#include "stp.h"


/* 	
 *	Port PORT is the port number associated to the socket address 
 *	used by a simulator.
 * 	Port PORT_TRASH is a Trash port , where a packet will be 
 *	losed because simulator will consider no ports attached to it 
 *	to listen.
*/

#define PORT 		2000
#define TRASH_PORT 	2001

/* 
 *	INIZIAL_PORT  must be defined as an initial port , 
 *	and it will determine a range
 * 	[(INIZIAL_PORT+1),(INIZIAL_PORT+NUM_BRIDGE)]
 * 	where NUM_BRIDGE will be user defined at start
 */
#define INITIAL_PORT 	1030

// static table allocated,  change this to simulate more bridges or lans

#define NUM_LAN 		20
#define NUM_BRIDGE 		40

/* 	
 *	Constant used to contain information about where to find configuration files 
 *	associated to every Bridge we are going to simulate.
 *	Create (mkdir) "bridgefile" somewhere , after that please insert configuration file
 *	as bridge.conf(n°) and then set STPFILE below to point at "stpfile" directory.
 * 	
 * 	Es.
 *   	Bridge with Identifier n°15 have :
			 bridge.conf15
		as a configuration file
 */
#define STPFILE 		"../bridgefile/bridge.conf"

/* 	
 *	The internet address associated with socket 
 *	is a localhost address plus port number, because simulation 
 *	on the same machine.
 */
void
init_sockaddr(struct sockaddr_in *name, uint16_t port)
{
	
	name->sin_family = AF_INET;
	name->sin_port = htons(port);
	name->sin_addr.s_addr = htonl(INADDR_LOOPBACK);  
}

// This function prepare a socket associated with "port" parameter 
int 
make_socket(uint16_t port)
{
	int sock;
	struct sockaddr_in name;
	if ((sock = socket(PF_INET,SOCK_DGRAM,0)) < 0 )
		{
			perror("Error creating a socket");
			exit(1);
		}
	//init_sockaddr(&name,HOST_NAME,port);
	name.sin_family = AF_INET;
	name.sin_port = htons(port);
	name.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
	if (bind(sock,(struct sockaddr *) &name, sizeof(name)) < 0)
		{
			perror("Bind error");
			exit(1);
		}
	return sock;
}

int main(int argc, char **argv)	
{	
	/* 	
	 *	lans and ports structure below are able to contain informations
	 *	on the Bridge LAN topology
	 */

	uint16_t lans[NUM_LAN][NUM_BRIDGE];
	uint16_t ports[NUM_LAN][NUM_BRIDGE];

	uint16_t lanto , numoflans , numofbridges ;
	
	char filename[60];
	char f[6];

	int val;
	FILE *fp;

	uint64_t bridgea;
	uint16_t bridgep,maxage,hellotime,forwardtime,numofports;
	uint16_t portnum,portp,lannum,pathcost,bridgenum,portNo;
	
	struct stp_configbpdu bpdu;

	struct	sockaddr_in addr;		

	int dg_sock;
	
	int nbpdur=0;
	int nbpdus=0;
	int i=0;
	int j=0;

	dg_sock = make_socket(PORT);
 
	/* 
	 *	Bridged LAN topology database initialization ( lans and ports 
	 *	structure initialization ).
	 */
	for(i=0; i<NUM_LAN; i++)
		{
			for(j=0; j<NUM_BRIDGE; j++)
				{	lans[i][j] = (uint16_t)(TRASH_PORT);
					ports[i][j] = 0;
					/* printf(" %hd %hd ",lans[i][j],ports[i][j]); */
				}
		}
	printf("\n");
	if (argc<3) {
		printf("Usage Error : lan_simulator [num of bridges]  [num of lans]\n");
		exit(1);
	}
	
	if (	(!(numofbridges = atoi(argv[1]))) 
			||
			(numofbridges > NUM_BRIDGE)
			||
			(numofbridges <1)
		) {
			printf("Usage Error : lan_simulator [num of bridges]  [num of lans]\n");
			exit(1);
	}
	
	if (	(!(numoflans = atoi(argv[2]))) 
			|| 
			(numoflans > NUM_LAN)
			|| 
			(numoflans < 1)
	 	) {
			printf("Usage Error : lan_simulator [num of bridges]  [num of lans]\n");
			exit(1);
	}
	 
	/* 	
	 *	This part scan configuration files in a directory specified 
	 *	in STPFILE constant	depending on max number of the bridges 
	 */
	bridgenum = 0;
	while(bridgenum<numofbridges)
			{
				printf("Bridge %i configuration file id :",bridgenum+1);
				fscanf(stdin,"%s",f);
				strcpy(filename,STPFILE);
				strlcat(filename,f,sizeof(filename));
				printf("SOCKET PORT OPENED %i\n",(INITIAL_PORT+(uint16_t)atoi(f)));
				if ((fp = fopen(filename,"r")) == NULL)
						{
						printf("Error loading bridge settings from file : %s\n",filename);
						exit(1);
						}
				if (fscanf(fp,"%lld %hd %hd %hd %hd %hd",&bridgea,&bridgep,&maxage,&hellotime,&forwardtime,&numofports) !=6)
					{	
						fclose(fp);
						exit(1);
					}
				portNo = 0;
				while(portNo<numofports)
					{
						if (fscanf(fp,"%hd %hd %hd %hd",&portnum,&lannum,&portp,&pathcost) != 4)
						{   
							fclose(fp);
							exit(1);
						}
						lans[lannum-1][bridgenum] = (uint16_t)(INITIAL_PORT+(uint16_t)atoi(f));
						ports[lannum-1][bridgenum] = portNo+1;
						portNo++;
					}
				fclose(fp);
				bridgenum++;
			}

	/* 	
	 *	Incoming BPDUs will be received on the socket dg_sock
	 *	and they are transmited , depending on the destination 
	 *	Bridge Identifier field, on the appropriated lans 
	 *	where the virtual bridges are listening.
	*/
	while (1) 
		{ 	
			// Listen waiting a BPDU packet ( coming from anywhere )
			if (read(dg_sock,&bpdu,sizeof(struct stp_configbpdu)) < 0 )
				{
					perror("Error reading bpdu from socket");
					exit(1);
				}
			nbpdur++;
		/* 	
		 *	Print information on the screen about BPDU received, BPDU transmitted, 
		 *	Bridge Identifier of Bridge transmitting BPDU, Port Identifier of Bridge transmitting BPDU, 
		 *	Bridge Identifier of Root Bridge stored in BPDU , and Root Path Cost stored in BPDU
		 */
		printf("Bpdus received %d Sends %d from %qd port %hd RootId %lld PCost %d\n",nbpdur,nbpdus,stp_get_address_bridgeid(bpdu.cbpdu_bridgeid),stp_get_portno_portid(bpdu.cbpdu_portid),stp_get_address_bridgeid(bpdu.cbpdu_designatedroot),bpdu.cbpdu_rootpathcost);
			fflush(stdout);
			
			//	BPDU are transmitted on the appropriatedd lan 
			lanto = stp_get_portno_portid(bpdu.cbpdu_portid);
			for(bridgenum=0; bridgenum<numofbridges; bridgenum++)
				{
					init_sockaddr(&addr,(uint16_t) lans[(lanto-1)][bridgenum]);
					stp_set_portno_portid(&bpdu.cbpdu_portid,ports[(lanto-1)][bridgenum]);
					if ( (sendto(dg_sock,&bpdu,sizeof(struct stp_configbpdu),MSG_OOB,(struct sockaddr *)&addr,sizeof(struct sockaddr_in))) != sizeof(struct stp_configbpdu) ) 
						{
							printf("Bpdu TCN received %i \n",nbpdur);
							nbpdus++;
						}
						else {
								nbpdus++;
						}
				}
			val = sizeof(struct stp_configbpdu);

			
		}
}
