/***************************************************************************
 *           virtual_bridge.c
 *
 *  Tue Oct  4 10:08:19 2005
 *  Copyright  2005  Francesco Carbone
 *  trini4mail@yahoo.it
 *	franci.carbone@gmail.com
 ****************************************************************************/



/* 	
 *	This code implements a program that will act as a Bridge in a simulated 
 *	Bridged Lan testing spanning tree protocol. A FreeBSD machine can implement 
 *	multiple logical bridge, each of them is made of a set of interfaces 
 *	identified by an address which is a 48 bit coded integer values.
 *	Bridge	is launched from a terminal spacifying the bridge address being 
 *	an integer value and setting an appropriated file as specified below.
 *	Information on the LANs where ports are attached and parameters to use 
 *	implementing the spanning tree protocol operations, are stored in a 
 *	configuration file as :
 *
 *	Bridge-Address Bridge-priority maxAge helloTime forwardDelayTime numOfPorts	
 *	Port-number LAN-connected Port-priority pathCost 							
 *		'	   '		'	  '					
 *		'	   '		'	  '						
 *	Port-number LAN-connected Port-priority pathCost
 *
 *	
 *	The module use only one socket to accept BPDUs from LANs to which it's 
 *	attached and trasmits all BPDUs to lan simulator through port LAN_SIM_PORT.
 *
 * 	The three greater tasks Bridge execute himself in order to implement 
 *	the protocol are :
 *	1)	Basic Operation of the Protocol 			( main process )
 *	2)	Management of the Protocol Entity			( _management_thread )
 *	3)	Management of the protocol Timers			( _timers_thread )
 *	
 *
 *	main process :
 *		1)	initalize semaphor variable for mutual exclusion data access
 *		2)	initialize variable to use starting threads
 *		3)	load informations stored in configuration files and save them 
 *			in "bridgeData" data structures
 *		4) 	start _timers_thread and _management_thread
 *		5)	listen for incoming BPDUs 
 *		6)	Operations of the spanning Tree protocol
 *		7)	print summury information on display
 *		8)	back to point 5)
 *	
 *	_timers_thread :
 *		1) 	initialize semaphor
 *		2) 	sleep 1 second
 *		3)	set semaphor lock
 *		4) 	use stp_tick(), declared in stp_bpdu.c, 
 *			function to manipulate timers
 *		5)	set semaphor unlock
 *		6)	return to point 2)
 *		
 *	_management_thread
 *		1)	initialize semaphor
 *		2)	wait input from keyboard
 *		3)	lock semaphor
 *		4)	execute operations in order to implement
 *			bridge management in respect of spanning
 *			tree protocol specifications
 *		5) 	unlock semaphor
 *		6)	retun to point 2
 *		
 *		This thread accept input from keyboard implementing Bridge Management,
 *		while Bridge don't stop executing operations of the spanning tree 
 *		protocol.
 *		Availble commands :
 *			h+enter		print help summury on the screen
 *			d+	"		input prompt to enter port number to disable
 *			e+	"		input prompt to enter port number to enable
 *			s+	"		sleep down bridge
 *			w+	"		wake up bridge
 *			p+  "		input prompt to enter port number and new priority
 *			b+ 	"		input prompt to enter bridge new priority
 *			c+	"		Clear screen
 *			i+	"		Refresh screen with Bridge info
 *			x+	"		Set Path Cost to a given port
 *
 *	THINGS TO REMEMBER
 *		- 	make sure INITIAL_PORT is equal as that specified 
 * 			in lan_simulator.c
 *		-	make sure LAN_SIM_PORT is equal as that specified in
 *			lan_simulator.c
 *	
 */

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <pthread.h>
#include <unistd.h>
#include <string.h>
#include "stp.h"
#include "stp_utility.h"


/* 	
 *	Replace this with your own hostname 
 *	This is used to bind a localhost address to socket
 * 	through bridge will connect to the bridged lan.
 */
#define HOST_NAME "trimobile.home.com"

/* 	Standard Directory and standard file name
 *	This is used to have a point on the directory  
 *	where must be placed configuration files with name bridge.conf(n°)
 *	examples: 	bridge.conf1 ---> file configuration bridge 1
 *				bridge.conf16 --> file configuration bridge 16
 */
#define STPFILE 		"../bridgefile/bridge.conf"

/* 
 *	Make sure this parameter is equal
 *	to the INITIAL_PORT defined in lan simulator
 *
 */
#define INITIAL_PORT 	1030

/* 	
 *	Lan_simulator running on the machine uses port LAN_SIM_PORT
 *	associated to localhost internet address linked to it.
 *	This is the unique socket opened by lan_simulator 
 *	and where every BPDU will be transmitted. The lan_simulator 
 *	uses the fields in BPDUs to decide on which LAN the packet 
 *	need to be transmitted.
 */
#define LAN_SIM_PORT 2000

/* 	
 *	Shared bridge data structures 
 *	See stp.h for more details about stp_bridgedata type
 */
struct stp_bridgedata *bridgeData;

// Semaphor to have mutual exclusion access to bridge data structures
pthread_mutex_t mut;


// Shared variables
/* 	
 *	They are used to have parameters printed 
 *	on the screen. This parameters are used to 
 *	get information on protocol performance.
 *	For example we can get information on communications
 *	bandwidth  consumed by the Bridges in enstablshing and maintaining 
 *	the spanning tree, looking at BPDUs number transmitted or received.
 */
int x = 0;

int port;

int dg_sock;

int bpdus_s = 0;
int bpdus_r = 0;

// Initialize a sockaddr_in structure with port value and AF_INET family
void
init_sockaddr(struct sockaddr_in *name, uint16_t port)
{
	name->sin_family = AF_INET;
	name->sin_port = htons(port);
	name->sin_addr.s_addr = htonl(INADDR_LOOPBACK);  
}


/*	
 *	Create a dgram socket associated to port "port"
 *	and return a socket number on success or 0
 */
int 
make_socket(uint16_t port)
{
	int sock;
	struct sockaddr_in name;
	if ((sock = socket(PF_INET,SOCK_DGRAM,0)) < 0 )
		{
			perror("Error creating a socket");
			exit(1);
		}
	name.sin_family = AF_INET;
	name.sin_port = htons(port);
	name.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
	if (bind(sock,(struct sockaddr *) &name, sizeof(name)) < 0)
		{
			perror("Bind error");
			exit(1);
		}
	return sock;
}


/* Clear Terminal Screen */
void
clrscr(int x)
{	int i;
	for(i=0;i<x;i++)
		printf("\n");
}
/* 	
 *	Send configuration BPDUs "configBpdu" through a port "portNo"
 *	Obviously this function is only for testing , and transmit all messages to 
 *	lan_simulator process, but it need to be implemented to transmit BPDUs 
 *	trought the link ( ports ) associated to the interfaces on the Bridge.
 */
void 
stp_send_config_bpdu(struct stp_bridgedata *bridgeData, uint16_t portNo, struct stp_configbpdu configBpdu)
{	
	uint16_t index = bridgeData->bd_ports[portNo]->pd_index;
	struct sockaddr_in addr;
	init_sockaddr(&addr,LAN_SIM_PORT);
	
	// Sending index for testing ( index  --> LAN n°index)
	stp_set_priority_portid(&index,stp_get_priority_portid(configBpdu.cbpdu_portid));
	configBpdu.cbpdu_portid=index;
	
	if ( (sendto(dg_sock,&configBpdu,sizeof(struct stp_configbpdu),MSG_OOB,( struct sockaddr *)&addr,sizeof(struct sockaddr_in))) != sizeof(struct stp_configbpdu) )
		printf("An error accoured during bpdu transmission\n");
	bpdus_s++;
}
 	
/* 	
 *	Thread used to manage timers on the Bridge.
 *	Because the three processes uses the same shared memory 
 *	( "bridgeData" Bridge Database ) a semaphor is used, 
 *	to ensure the correct access.
 *	The only function of this thread is to call "stp_tick()" every second.
 *	This function specified in "stp_bpdu.c" module, manage timers used by 
 *	the Bridge and manipulate Bridge parameters as specfied in the 
 *	operations of the protocol.
 *
 *	Informations about Bridge ports status are displayed on the screen
 *	every 1 second as (x,yy,z) where :
 *
 *			x 	is a port number
 *			yy 	indicates if the port is Root Port or Designated Port
 *			z 	indicates Port State
 *
 *	For example :	
 *			(1,DP,F)
 *			Port 1 is in Forwarding State and it is a
 *			Designated Port for the lan
 *					
 *			(5,RP,F) 	Port 5 is in Forwarding State and it is 
  *						Root Port 
 */
void *
_timer_thread()
{	
	int i=0;
	
	pthread_mutex_init(&mut,NULL);
	
	while (1) {
		
		sleep(1);
		pthread_mutex_lock(&mut);
		stp_tick(bridgeData);
		pthread_mutex_unlock(&mut);
	
		/*	Printing on the screen, port parameters
			every second following port data changes
		*/
		printf(" ");
		for(i=1; i<=bridgeData->bd_numofports; i++)
			{
				printf("(%hd,",stp_get_portno_portid(bridgeData->bd_ports[i]->pd_portid));
				if (stp_designated_port(bridgeData,i)==1)		
					{
					printf("DP,"); 	// DP = designated port meaning that port is a designated port for the lan
					}
				else	if (stp_root_port(bridgeData,i)==1) {	
							printf("RP,");
							}
						else	{ printf("--,");
								}
						
				switch (bridgeData->bd_ports[i]->pd_state)	{	
					case 0: printf("D) "); // port is disabled
						break;
					case 4:	printf("B) "); // port is in blocking state
						break;
					case 1:	printf("L) "); // port is in listening state
						break;
					case 2:	printf("l) "); // port is in learning state
						break;
					case 3:	printf("F) "); // port is in forwarding state
						break;
					default: break;
				}
			}
		/*
		 *	HT = Hello Timer 
		 *	BR = BPDU received
		 *	BS = BPDU transmitted
		 *	Root = Bridge address identifying the Designated Root
		 *
		 */
		printf("HT %hd BR %d BS %d Root %lld\15",stp_get_timer_value(bridgeData->bd_timerhello),\
			bpdus_r,bpdus_s,stp_get_address_bridgeid(bridgeData->bd_designatedroot));
		fflush(stdout);
		
	}
}


/*
 *	Bridge Management is implemented on the Bridge
 *	A Library functions defined in "MANAGEMENT OF THE
 *	BRIDGE PROTOCOL ENTITY" section of stp_bpdu.c are used.
 *
 *	TYPE h+enter to see available commands on the terminal
 *	window
 */
void *
_management_thread()
{	
	int portnum=0;
	int pNum=0;
	uint16_t p;
	uint32_t cost;
	
	pthread_mutex_init(&mut,NULL);
	
	
	/*
	 *	Bridge and Port parameters can be modified through
 	 *	Bridge Management by using command below.	
	 *	Input command described can modify the active topology 
	 *	of the Bridged LAN by setting new port and bridge priority
	 *	ecc.
	 */
	while (1) {
		printf(" TYPE h+enter FOR HELP or i+enter to refresh info\n");
		switch getchar() {
			case	'h':	printf("Availble commands :\n\
	h+enter		print help summury on the terminal window\n\
 	d+enter		input prompt to enter port number to disable\n\
 	e+enter		input prompt to enter port number to enable\n\
 	s+enter		sleep down bridge\n\
 	w+enter		wake up bridge\n\
 	p+enter		input prompt to enter port number and new priority\n\
 	b+enter		input prompt to enter bridge new priority\n\
	c+enter		Clear terminal\n\
 	i+enter		Refresh window terminal  with Bridge info\n\
	x+enter		Set Path Cost to a given Port\n");
							break;
			/*	
			 * Disabling port , interface linked to it will not
			 * operate in active topology anymore
			 *
			 */
			case	'd':	printf("\n Insert port number to disable :\n");
							fscanf(stdin,"%i",&portnum);
							if (portnum<=bridgeData->bd_numofports) {
								pthread_mutex_lock(&mut);
								stp_disable_port(bridgeData, portnum);
								pthread_mutex_unlock(&mut);
							}
							break;
			/*	
			 * Enabling port , interface linked to it will
			 * operate in active topology again
			 *
			 */
			case	'e':	printf("\n Insert Port Number to enable : \n");
							fscanf(stdin,"%i",&portnum);
							if (portnum<=bridgeData->bd_numofports) {
								pthread_mutex_lock(&mut);
								stp_enable_port(bridgeData, portnum);
								pthread_mutex_unlock(&mut);
							}
							break;
			/*
			 *	Bridge will be stopped setting mutex variable mut
			 *	locked
		 	 *	With w command turn on the Bridge again setting
			 *	mut unlocked
		     */
			case	's':	printf("BRIDGE DOWN\n");
							pthread_mutex_lock(&mut);
							break;
			case	'w':	printf("BRIDGE UP");
							pthread_mutex_unlock(&mut);
							break;
			/*
			 *	Input port will be modified with a new priority
			 *	coded in a 8 more significant bits of the 16 
			 *	bits port identifier.
			 *	Bridge is not stopped executing spanning tree 
 			 *	operations
			 */
			case	'p':	printf("\n Insert port number :\n");
							fscanf(stdin,"%i",&portnum);
							if (portnum<=bridgeData->bd_numofports) {
								printf(" Insert New PRIORITY : \n");
								fscanf(stdin,"%hd",&p);
								pthread_mutex_lock(&mut);
								stp_set_port_priority(bridgeData, portnum,p);
								pthread_mutex_unlock(&mut);
							}
							break;
			/*
			 *	Bridge priority coded in a 16 more significant bits 
			 *	of Bridge Identifier will be modified with a new p
			 */							
			case	'b':	printf("\n Insert new Bridge Priority :\n");
							fscanf(stdin,"%hd",&p);
							pthread_mutex_lock(&mut);
							stp_set_bridge_priority(bridgeData,p);
							pthread_mutex_unlock(&mut);
							break;
			case	'c':	clrscr(50);
							break;
			case	'i':	clrscr(50);
							printf("Bridge A %lld Bridge P %hd B_maxAge %hd B_helloTime %hd  B_Forward_D %hd B_numports %hd\n",\
									stp_get_address_bridgeid(bridgeData->bd_bridgeid),stp_get_priority_bridgeid(bridgeData->bd_bridgeid),bridgeData->bd_bridgemaxage,\
									bridgeData->bd_bridgehellotime,bridgeData->bd_bridgeforwarddelaytime,\
									bridgeData->bd_numofports);
							pNum=1;
							while ( pNum <= (bridgeData->bd_numofports) )
							{	printf("Port %hd Cost %hd PortNum %hd PortPrior %hd\n",(int)bridgeData->bd_ports[pNum]->pd_portid,\
										(int)bridgeData->bd_ports[pNum]->pd_pathcost,\
										(int)stp_get_portno_portid(bridgeData->bd_ports[pNum]->pd_portid),\
										(int)stp_get_priority_portid(bridgeData->bd_ports[pNum]->pd_portid));
								pNum++;
							}
							/*	Printing on display bridge parameters loaded from conf. file */
							for(x=1; x<=bridgeData->bd_numofports; x++)
								printf(" (%hd)%hd  ",stp_get_portno_portid(bridgeData->bd_ports[x]->pd_portid),bridgeData->bd_ports[x]->pd_index);
							printf("\n");
							break;
			case	'x':printf("\n Insert port number :\n");
							fscanf(stdin,"%i",&portnum);
							if (portnum<=bridgeData->bd_numofports) {
								printf(" Insert new Path Cost : \n");
								fscanf(stdin,"%d",&cost);
								pthread_mutex_lock(&mut);
								stp_set_path_cost(bridgeData,portnum,cost);
								pthread_mutex_unlock(&mut);
							}
							break;
			default:	break;
		}
	}
}


int main(int argc,char **argv)
{	

//	int arg;
	int x=0;
	int z=0;
	char filename[60];

	uint16_t portNo;

	/*	Confifuration Bpdu data type defined in "stp.h" file */
	struct stp_configbpdu bpdu;

	pthread_attr_t 	attr;
	pthread_t 		timer_threadid;  	// _timer_thread identifier
	pthread_t		management_threadid;// _management_thread identifier
	
	if (argc<2) {
		printf("Usage Error : virtual_bridge [bridge file id]\n");
		exit(1);
	}
	if (!(port = atoi(argv[1]))) {
		printf("Usage Error : virtual_bridge [bridge file id]\n");
		exit(1);
	}
	// Semaphor initialization	
	pthread_mutex_init(&mut,NULL);
	pthread_mutex_lock(&mut);
	
	
	// dg_sock will contain socket number associated to the bridge
	dg_sock = make_socket(INITIAL_PORT+port);
	printf("SOCKET PORT OPENED %i\n",(INITIAL_PORT+port)); 
	
	/* 	
	 *	Load data stored in the "filename" configuration file
	 *	and save it in the "bridgeData" data structures, where 
	 *	bridge parameters are stored.
	 */	
	strcpy(filename,STPFILE);
	strlcat(filename,argv[1],sizeof(filename));
	if ((bridgeData = stp_create_bridge(filename)) == NULL)
		{
			printf("Error loading bridge settings from file : %s\n",filename);
			exit(1);
		}
	
	/* Starting _timer_pthread */
	x = pthread_create(&timer_threadid,0,_timer_thread,0); 
	x = pthread_attr_init(&attr);
		
	/* Starting _management_pthread */
	z = pthread_create(&management_threadid,0,_management_thread,0); 
	z = pthread_attr_init(&attr);
	
	stp_initialization(bridgeData);
	
	pthread_mutex_unlock(&mut);
	
	/*	Printing on display bridge parameters loaded from conf. file */
	for(x=1; x<=bridgeData->bd_numofports; x++)
		printf(" (%hd)%hd  ",stp_get_portno_portid(bridgeData->bd_ports[x]->pd_portid),bridgeData->bd_ports[x]->pd_index);
	printf("\n");
	while (1) 
		{ 	
			/* Listening for BPDUs on socket "dg_sock" */ 
			if (read(dg_sock,&bpdu,sizeof(struct stp_configbpdu)) < 0 )
				{
					perror("Error reading bpdu from socket");
					exit(1);
				}
			bpdus_r++;
			
			/*	
			 *	Bridge control the type of BPDU received and accept him
			 *	if the bridge destination identifier field is equal to bridge 
			 *	identifier stored in "bridgeData" , identifying the bridge himself.
			 */
			if (bpdu.cbpdu_bpdutype == CONFIG_BPDU_TYPE)
				{
					if (!((uint32_t)bpdu.cbpdu_bridgeid == (uint64_t)bridgeData->bd_bridgeid))
						{	
							pthread_mutex_lock(&mut);
							portNo = stp_get_portno_portid(bpdu.cbpdu_portid);
							stp_received_configuration_bpdu(bridgeData,portNo,bpdu);
							pthread_mutex_unlock(&mut);
						}
				}
			else 
				{
					if (!((uint32_t)bpdu.cbpdu_bridgeid == (uint64_t)bridgeData->bd_bridgeid))
						{	
							pthread_mutex_lock(&mut);
							portNo = stp_get_portno_portid(bpdu.cbpdu_portid);
							stp_received_tcn_bpdu(bridgeData,portNo);
							pthread_mutex_unlock(&mut);
						}
				}
			
		}
	
}
